{% extends 'hidp/base_post_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Manage linked trusted services' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Manage your linked trusted services' %}</h1>

  <p>
    {% blocktranslate trimmed with provider=provider.name %}
      Successfully linked your {{ provider }} account.
    {% endblocktranslate %}
  </p>

  {% include 'hidp/includes/forms/submit_row.html' with cancel_label=_('Back') cancel_url=back_url %}
{% endblock %}
