{% extends 'hidp/base.html' %}
{% comment %}
This template serves as the base template for all pages that are displayed before a user is logged-in.

This includes the following pages (non-exhaustive list):

- Registration flow
- Email verification flow
- Login flow
- OTP verification flow
- Password reset flow
- Terms of service
- Logout confirmation

It extends the base.html template and is an entrypoint for customizing the layout of the pre-login pages.
{% endcomment %}
