# Generated by Django 5.1.1 on 2024-10-03 10:24

import django.db.models.deletion
import hidp.compat.uuid7
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailChangeRequest',
            fields=[
                ('id', models.UUIDField(default=hidp.compat.uuid7.uuid7, editable=False, primary_key=True, serialize=False)),
                ('proposed_email', models.EmailField(max_length=254, verbose_name='proposed email address')),
                ('current_email', models.EmailField(max_length=254, verbose_name='current email address')),
                ('confirmed_by_proposed_email', models.BooleanField(default=False, verbose_name='confirmed by proposed email')),
                ('confirmed_by_current_email', models.BooleanField(default=False, verbose_name='confirmed by current email')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('last_modified', models.DateTimeField(auto_now=True, verbose_name='last modified')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='email_change_request', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'email change request',
                'verbose_name_plural': 'email change requests',
            },
        ),
    ]
