{% extends 'hidp/base_pre_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Verify email' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Verify email' %}</h1>

  {% if form.first_name and form.last_name %}
    <p>{% translate 'To continue, please fill in your first and last name before verifying your email.' %}</p>
  {% else %}
    <p>{% translate 'To continue, please verify your email.' %}</p>
  {% endif %}

  <form method="post">
    {% csrf_token %}
    {{ form }}
    {% include 'hidp/includes/forms/submit_row.html' with submit_label=_('Verify email') %}
  </form>
{% endblock %}
