{% extends 'hidp/base_pre_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Reset password' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Reset password' %}</h1>

  <p>
    {% blocktranslate trimmed %}
      Forgotten your password? Enter your email address to receive a link to reset your password.
    {% endblocktranslate %}
  </p>

  <form method="post">
    {% csrf_token %}
    {{ form }}
    {% include 'hidp/includes/forms/submit_row.html' with submit_label=_('Continue') cancel_url=cancel_url cancel_label=_('Back') %}
  </form>
{% endblock %}
