{% extends 'hidp/base_post_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Cancel email change request' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Cancel email change request' %}</h1>

  <p>{% translate 'The request to change your account email address has been successfully canceled.' %}</p>

  {% include 'hidp/includes/forms/submit_row.html' with cancel_label=_('Back') cancel_url=back_url %}
{% endblock %}
