{% extends 'hidp/base_pre_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Reset password' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Reset password' %}</h1>

  <p>{% translate 'Please enter your new password twice to ensure you typed it correctly.' %}</p>

  <form method="post">
    {% csrf_token %}
    <!-- Hidden text input to assist password managers with updating the stored credentials -->
    <input hidden autocomplete="username" value="{{ form.user.get_username }}">
    {{ form }}
    {% include 'hidp/includes/forms/submit_row.html' with submit_label=_('Change password') %}
  </form>
{% endblock %}
