{% extends 'hidp/base_pre_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Verification completed' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Verification completed' %}</h1>

  <p>
    {% translate 'Your email address is now verified.' %}
    {% blocktranslate %}You can now <a href="{{ login_url }}">sign in</a>.{% endblocktranslate %}
  </p>
{% endblock %}
