{% extends 'hidp/base_pre_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Reset password' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Reset password' %}</h1>

  <p>{% translate 'Please check your email for a link to reset your password.' %}</p>

  <p>{% translate 'It may take a few minutes for this email to arrive. If you do not see it, check your spam folder. If you still cannot find it, make sure you entered the correct email address.' %}<p>
{% endblock %}
