{% extends 'hidp/base_pre_login.html' %}
{% load i18n %}

{% block title %}{% translate 'Reset password' %}{% endblock %}

{% block main %}
  <h1>{% translate 'Reset password' %}</h1>

  <p>{% translate 'Your password has been set. You can now sign in using your new password.' %}</p>

  <p><a href="{{ login_url }}">{% translate 'Sign in' %}</a></p>
{% endblock %}
