<!doctype html>
<html>
<head>
  <meta charset="utf-8">
  <title>{% block title %}{% endblock %}</title>
</head>
<body>
{% block body %}
  {% block main %}{# Main application content #}{% endblock %}
{% endblock %}
</body>
</html>
