{% extends "hidp/base_email.html" %}
{% load i18n %}

{% block content %}
  <p>{% translate 'Your password has been changed.' %}</p>
  <p>{% translate 'If you did not change your password, please reset your password immediately using the following link:' %}</p>
  <p><a href="{{ password_reset_url }}">{{ password_reset_url }}</a></p>
{% endblock %}
