{% extends "hidp/base_email.html" %}
{% load i18n %}

{% block content %}
  <p>{% translate 'You requested a link to reset your password.' %}</p>
  <p>{% translate 'However, your account does not currently have a password set.' %}</p>
  <p>{% translate 'Use the following link to set a password:' %}</p>
  <p><a href="{{ set_password_url }}">{{ set_password_url }}</a></p>
{% endblock %}
