{% extends "hidp/base_email.html" %}
{% load i18n %}

{% block content %}
<p>
  {% blocktranslate trimmed %}
  A recovery code was used to log in to your account.

  If you did not attempt to log in, please reset your password immediately.
  {% endblocktranslate %}
</p>
<p>
  {% blocktranslate trimmed count tokens=recovery_codes_count %}
  You have {{ tokens }} recovery code left.
  {% plural %}
  You have {{ tokens }} recovery codes left.
  {% endblocktranslate %}

  {% blocktranslate trimmed %}
  You can generate new recovery codes by visiting the two-factor authentication settings page:
  {% endblocktranslate %}
</p>
<p><a href="{{ otp_management_url }}">{{ otp_management_url }}</a></p>
{% endblock %}
