{% load i18n %}
{% if providers %}
  <ul>
    {% for item in providers %}
      <li>
        {% include 'hidp/includes/federated/oidc_provider_list_option.html' %}
      </li>
    {% endfor %}
  </ul>
{% endif %}
