{% extends 'hidp/base.html' %}
{% comment %}
This template serves as the base template for all pages that are displayed after a user is logged-in.

This includes the following pages (non-exhaustive list):

- Profile management
- Password change flow
- OIDC account (un)linking flow
- Email change flow

It extends the base.html template and is an entrypoint for customizing the layout of the post-login pages.
{% endcomment %}
