{% extends "hidp/base_email.html" %}
{% load i18n %}

{% block content %}
  <p>{% translate 'Use the following link to reset your password:' %}</p>
  <p><a href="{{ password_reset_url }}">{{ password_reset_url }}</a></p>
{% endblock %}
